#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <libssp.h>
#include <libsspprof.h>
#include <wiringPi.h>
#include <signal.h>
#include<unistd.h>
#include <sys/time.h>
#include <omp.h>
#include<opencv/cv.h>
#include <opencv/highgui.h>
// sudo apt-get install libopencv-dev
// sudo apt-get install libgl1-mesa-dri
#include <jpeglib.h>
// sudo apt-get install libjpeg-dev
#include "../common/isreg.h"

#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE
#define FRAME_FIFO_SIZE 2

int ProcKey(int KeyCode,struct ssp_handle *handle);
int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

struct isreg *isregp;
static int frame_count;

volatile int stop_process;
struct ssp_profile *profile;
IplImage *BGRImg, *BayerImg;

static void signal_handler(int sig)
{
  printf("Terminated.\n");
  if(sig == SIGINT)
    stop_process = 1;
}


void frame_drop_func(struct ssp_handle *handle){

  printf("Frame dropped by camera.\n");

}

void frame_drop_preprocess_func(struct ssp_handle *handle){

  printf("Frame dropped at preprocess.\n");
}

unsigned char *img;
char *filename;
int WaitCount;
int ExpTimePlan, ExpTime;
size_t size = 0;

int RegAdr_Gain;
int RegVal_Gain;
int ImgWidth, ImgHeight;
int WaitCount;
int ExpStatus;
char str[100];
CvFont font;
#define EXPOSE 0
#define PAUSE 1

void fame_preprocess(struct ssp_handle *handle, struct ssp_frame *frame){
  //  printf("Preprocess is active. frame_count = %d\n", frame_count);

//  int i;
  size_t i;

  usleep(10000);

  unsigned char *frame_data = ssp_get_frame_data(frame);

  if(ExpStatus == EXPOSE){
    frame_count ++;
    if(frame_count % 1 == 0) printf("frame_count = %d\n", frame_count);

    if(frame_count == 1){
      ssp_get_frame_size(frame, &size);
      img = (unsigned char *) malloc(sizeof(unsigned char) * 2 * size);
      #pragma omp parallel for
      for(i = 0; i < size; i++){
        img[i] =  0;
      }
    }else if(frame_count > FRAME_FIFO_SIZE){
      ExpTime++;
      #pragma omp parallel for
      for(i = 0; i < size; i+=2){
        unsigned int accum = (unsigned int)img[i] + (unsigned int)img[i+1] * 0x100;
        unsigned int frame = (unsigned int)frame_data[i] + (unsigned int)frame_data[i+1] * 0x100;
        unsigned int tmp = (frame > 64) ? accum  + frame - 64 : accum;
    
//        if(tmp > 0x3ff){
        if(tmp > 0xffff){
//          img[i+1] = 0x03;
          img[i+1] = 0xff;
          img[i]   = 0xff;
        }else{
          img[i+1] = (unsigned char)(tmp / 0x100);
          img[i]   = (unsigned char)(tmp & 0xff);
        }
        tmp = tmp >> 2;
        (BayerImg->imageData)[i/2] = (unsigned char)(tmp > 0xff ? 0xff : tmp);
      }
    }
  }else{
      #pragma omp parallel for
      for(i = 0; i < size; i+=2){
        (BayerImg->imageData)[i/2] = (img[i+1] * 0x100 + img[i]) >> 2;
      }
  }

  cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);
  ssp_release_frame(frame);
  sprintf(str,"Gain:%d Expose:%dsec(%dsec)", RegVal_Gain, ExpTime, ExpTimePlan);
  cvPutText(BGRImg, str, cvPoint(1+10*profile->ImageProperty.Width/3200, 150*profile->ImageProperty.Height/2400), &font, CV_RGB(200,200,200));
  cvShowImage("SSP Viewer", BGRImg);
  int KeyCode = cvWaitKey(2);
  if(KeyCode == 'q'){
    WaitCount = frame_count;
    printf("Quit\n");
  }else if(KeyCode == 'p'){
    ExpStatus = PAUSE;
    printf("Pause\n");
  }else if(KeyCode == 's'){
    ExpStatus = EXPOSE;
    printf("Start\n");
  }else if(KeyCode == 'u'){
    ExpTimePlan++;
    WaitCount++;
    printf("Extend to %d\n", ExpTimePlan);
  }else if(KeyCode == 'd'){
    ExpTimePlan--;
    WaitCount--;
    printf("Shorten to %d\n", ExpTimePlan);
  }

  if(frame_count == WaitCount){
    stop_process = 1;
    FILE *file;
    file = fopen(filename, "wb");
    if(file){
      fwrite(img, size, 1, file);
      fclose(file);
      printf("Saved RAW image to %s\n", filename);
    }
  }

}

#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE
#define REGADR_GAIN_IMX219 0x0157
#define REGADR_GAIN_IMX408 0x0205

int main(int argc, char **argv){

//  if(argc < 3){
  if(argc < 6){
    printf("SSP library still sample program in RAW16\n");
    printf("Copyright (c) Streamtechnology Corporation.\n");
    printf("              http://www.streamtechnology.co.jp/\n");
//    printf("Usage: %s profile_file_name output_file_name\n", argv[0]);
    printf("Usage: %s profile_file_name model_name output_file_name Gain Exposure\n", argv[0]);

    return 0;
  }

  char *profile_filename;
  char *model_name;

  profile_filename = argv[1];
  model_name = argv[2];
  filename = argv[3];

  RegVal_Gain = atoi(argv[4]);

  ExpTimePlan   = atoi(argv[5]);
  WaitCount = ExpTimePlan + FRAME_FIFO_SIZE;  //FIFOɗ܂f[^j

  isregp = (struct isreg *)check_name(model_name);
  if(isregp == NULL){
    printf("not found name %s\n",model_name);
    return(1);
  }

  stop_process = 0;
  if (signal(SIGINT, signal_handler) == SIG_ERR)
    return 2;
  
  struct ssp_settings settings;

  frame_count = 0;
  ExpTime = 0;
  
  // SSP settings
  settings.lib_settings.NumFrameFIFOSize = FRAME_FIFO_SIZE;
  settings.lib_settings.NumThreadForBuildInPreprocess = 1;
  settings.lib_settings.frame_drop_cam_user_func = frame_drop_func;
  settings.lib_settings.frame_drop_pre_user_func = frame_drop_preprocess_func;
  settings.lib_settings.fame_preprocess_user_func = fame_preprocess;
  settings.frame_preprocess_options.FormatConvertType = SSP_FRAME_BAYER16;
  settings.frame_preprocess_options.GammaCorrectionEnable = SSP_DISABLE;
  settings.frame_preprocess_options.GammaVal = 2.2;
  settings.camera_settings.PowerOnResetUsec = 1000000;

  struct ssp_handle *handle;

  ssp_read_profile(profile_filename, &profile);

  printf("Setting summary: %d x %d @ %d fps %d bits\n",
	 profile->ImageProperty.Width,
	 profile->ImageProperty.Height,
	 profile->ImageProperty.Frequency,
	 profile->ImageProperty.BayerBits);

  // OpenCV Settings
  cvNamedWindow("SSP Viewer", CV_WINDOW_KEEPRATIO); 
  BayerImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,1);
  BGRImg = cvCreateImage(cvSize(profile->ImageProperty.Width  , profile->ImageProperty.Height),IPL_DEPTH_8U,3);
  cvInitFont(&font, CV_FONT_HERSHEY_SIMPLEX, 5.0*profile->ImageProperty.Width/3200, 5.0*profile->ImageProperty.Height/2400, 0, 10*profile->ImageProperty.Width/3200, CV_AA);

  int i,j;
  for(i=0;i<1;i++){ // for repeat initialization test.
    int retval;
    if((retval = ssp_initialize(&handle, profile, &settings))!=SSP_SUCCESS){
      
      printf("Can not initialize ssp library. %x\n",retval);
      return 0;
    }

    //Analog Gain set
    __CCIRegWrite(handle, isregp->Adr_GainL, RegVal_Gain & 0xFF);
    if(isregp->Adr_GainH != -1){
      __CCIRegWrite(handle, isregp->Adr_GainH, (RegVal_Gain>>8)&0x1);
    }
    printf("Gain value = %d\n",RegVal_Gain);

    printf("Initialize done.\n");
    printf("start streaming\n");
    if((retval = ssp_start_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }

    while(frame_count < WaitCount){
      if(stop_process != 0) break;
    }

    printf("FrameCount = %d\n", frame_count);
    printf("stop streaming\n");
    if((retval = ssp_stop_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }
    
    printf("Sync Wait.\n");
    ssp_sync_event(handle);
    printf("Sync Done.\n");

    if(ssp_finalize(handle) != SSP_SUCCESS){
      printf("Error during finalization.\n");
      return 0;
    }

    cvReleaseImage(&BayerImg);
    cvReleaseImage(&BGRImg);
    cvDestroyWindow("SSP Viewer");

    printf("Finalize done.\n");fflush(stdout);
  }
  return 1;
}
