#include <stdio.h>
#include <string.h>
#include <glob.h>
#include <omp.h>
#include <opencv/cv.h>
#include <opencv/highgui.h>

// Key Code
#define SHIFT 0x01
#define CTRL  0x04
#define ALT   0x08
#define RIGHT 0x51
#define UP    0x52
#define LEFT  0x53
#define DOWN  0x54
#define ESC   0x1B

// 16bit raw Image Size
#define ImgSize Width*Height

//Search Area
#define MARGIN 256

int Width, Height;
int zWidth, zHeight;  //Zoom Window
int tWidth, tHeight;  //Template Window
int sWidth, sHeight;  //Search Window
int FrameNo    = 0;
int FrameCount = 0;
int ZoomMode   = 0;
int TempMode   = 0;
IplImage *BGRImg, *BayerImg, *ZoomImg;                  // for Display
IplImage *ReadBGRImg, *ReadBayerImg, *TempBGRImg;       // for Display
IplImage *SrchBGRImg;                                   // for Display
unsigned short *AccImg, *ReadImg, *DarkImg, *WriteImg;  // for Raw Data
char *dark_filename;
char str[100];
char wfilename[100];
char stop_process = 0;
CvFont font;
CvPoint from, to;   //for Zoom & Template Area
CvPoint tpos;       //Template Position
CvPoint offset;     //Image Offset for Stacking
CvPoint sfrom, sto; //for Search Area


void imageCopy(int x, int y, IplImage *SrcImg, IplImage *DstImg){
  int i,j,k,l;
  for(j=y, l=0; l < DstImg->height; j++, l++){
    for(i=x, k=0; k < DstImg->width; i++, k++){
      (DstImg->imageData)[DstImg->width * 3 * l + k * 3    ] = (SrcImg->imageData)[SrcImg->width * 3 * j + i * 3    ];
      (DstImg->imageData)[DstImg->width * 3 * l + k * 3 + 1] = (SrcImg->imageData)[SrcImg->width * 3 * j + i * 3 + 1];
      (DstImg->imageData)[DstImg->width * 3 * l + k * 3 + 2] = (SrcImg->imageData)[SrcImg->width * 3 * j + i * 3 + 2];
    }
  }
}


void onMouse(int event, int x, int y, int flags, void* param){
  if(!strcmp(param, "Read")){
  }else if(!strcmp(param, "Work")){
    if(event== CV_EVENT_LBUTTONDOWN){
      printf("Zoom     Area: %4d,%4d", x, y);
      from = cvPoint(x, y);
      from.x = from.x/2*2;
      from.y = from.y/2*2;
    }else if(event== CV_EVENT_LBUTTONUP){
      printf(" - %4d,%4d", x, y);
      to = cvPoint(x, y);

      if(to.x - from.x < 0){
        to.x = from.x;
        from.x = x;
      }
      if(to.y - from.y < 0){
        to.y = from.y;
        from.y = y;
      }
      zWidth  = (to.x - from.x)/4*4;
      zHeight = (to.y - from.y)/4*4;
      to.x = from.x + zWidth;
      to.y = from.y + zHeight;
      printf(" (%4d x %4d)\n",zWidth, zHeight);

      if(zWidth == 0 || zHeight== 0)
        return;

      cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);
      ZoomImg = cvCreateImage(cvSize(zWidth, zHeight),IPL_DEPTH_8U,3);
      ZoomMode   = 1;
      imageCopy(from.x, from.y, BGRImg, ZoomImg);
      cvNamedWindow("Zoom", CV_WINDOW_KEEPRATIO); 
      cvShowImage("Zoom", ZoomImg);

      if(TempMode == 0){
        sfrom.x = from.x - MARGIN > 0 ? from.x - MARGIN : 0;
        sfrom.y = from.y - MARGIN > 0 ? from.y - MARGIN : 0;
        sto.x = to.x + MARGIN < Width  ? to.x + MARGIN : Width;
        sto.y = to.y + MARGIN < Height ? to.y + MARGIN : Height;
        sWidth = sto.x - sfrom.x;
        sHeight= sto.y - sfrom.y;
        tpos.x = from.x - sfrom.x;
        tpos.y = from.y - sfrom.y;
        tWidth  = zWidth;
        tHeight = zHeight;
        printf("Template Area: %4d,%4d - %04d,%04d (%4d x %4d)\n",  from.x,  from.y,  to.x,  to.y, tWidth, tHeight);
        printf("Search   Area: %4d,%4d - %04d,%04d (%4d x %4d)\n", sfrom.x, sfrom.y, sto.x, sto.y, sWidth, sHeight);

        TempBGRImg = cvCreateImage(cvSize(tWidth, tHeight),IPL_DEPTH_8U,3);
        imageCopy( from.x,  from.y, BGRImg, TempBGRImg);
        cvNamedWindow("Tenmpate", CV_WINDOW_KEEPRATIO); 
        cvShowImage("Tenmpate", TempBGRImg);

        SrchBGRImg = cvCreateImage(cvSize(sWidth, sHeight),IPL_DEPTH_8U,3);
        imageCopy(sfrom.x, sfrom.y, BGRImg, SrchBGRImg);
        cvNamedWindow("SearchArea", CV_WINDOW_KEEPRATIO); 
        cvShowImage("SearchArea", SrchBGRImg);

        TempMode   = 1;
      }

      cvRectangle(BGRImg, cvPoint(from.x,  from.y),  cvPoint(to.x,  to.y), CV_RGB(200,255,255), 2, 10, 0);
      cvRectangle(BGRImg, cvPoint(sfrom.x, sfrom.y), cvPoint(sto.x, sto.y), CV_RGB(255,0,0), 2, 10, 0);
      cvShowImage("Work", BGRImg);
    }
  }
}


void SaveImage(int FrameCount){
    sprintf(wfilename,"RawStack_%04d.raw", FrameCount);

    FILE *file;
    file = fopen(wfilename, "wb");
    if(file){
      fwrite(WriteImg, ImgSize, 2, file);
      fclose(file);
      printf("Saved RAW image to %s\n", wfilename);
    }
}

#define THRESHOLD 0.7
#define OFFSETZERO 1
#define NOTFOUND 2

int AutoDetect(IplImage *SrcImg, IplImage *TemplateImg){
  IplImage *ResultImg;
  double max_val;
  CvPoint max_pos;

  if(TempMode == 1){
    CvSize result_size = cvSize(SrcImg->width - TemplateImg->width + 1, SrcImg->height - TemplateImg->height + 1);
    ResultImg = cvCreateImage(result_size, IPL_DEPTH_32F, 1);

    cvMatchTemplate( SrcImg, TemplateImg, ResultImg, CV_TM_CCORR_NORMED );
    cvMinMaxLoc(ResultImg, NULL, &max_val, NULL, &max_pos, NULL);
    cvReleaseImage(&ResultImg);

    printf("  maxval:%lf ",max_val);
    if (max_val > THRESHOLD){
      offset.x = (tpos.x - max_pos.x)/2*2;
      offset.y = (tpos.y - max_pos.y)/2*2;
      printf("Match Position:%4d,%4d Offset:%4d,%4d\n", max_pos.x, max_pos.y, offset.x, offset.y);
      if(offset.x == 0 && offset.y == 0){
        return OFFSETZERO;
      }
    }else{
      printf("not found match\n");
      return NOTFOUND;
    }
  }
  return  0;
}


void ReadImage(char *rfilename){
  size_t i;
  unsigned int readi;
  unsigned int accum;
  unsigned int tmp;
  int err_code;
  char *chp;

  offset.x = 0;
  offset.y = 0;

  FILE *file;
  file = fopen(rfilename, "rb");
  if(file){
    fread(ReadImg, ImgSize, 2, file);
    fclose(file);
    printf("%s Read\n", rfilename);
  }else{
    printf("Can't read Image File.\n");
  }
  #pragma omp parallel for
  for(i = 0; i < ImgSize; i++){
    unsigned int readi = (unsigned int)ReadImg[i];
    unsigned int darkv = (unsigned int)DarkImg[i];
    unsigned int tmp   = (readi > darkv) ? readi - darkv : 0;
    ReadImg[i]   = (unsigned short)tmp;
    tmp = tmp >> 2;
    (ReadBayerImg->imageData)[i] = (unsigned char)(tmp > 0xff ? 0xff : tmp);
  }
  cvCvtColor(ReadBayerImg, ReadBGRImg, CV_BayerBG2BGR);
  cvCopy(ReadBGRImg, BGRImg, NULL);
  if(TempMode == 1)
    cvRectangle(BGRImg, cvPoint(sfrom.x, sfrom.y), cvPoint(sto.x, sto.y), CV_RGB(255,0,0), 2, 10, 0);
  chp = strrchr(rfilename,'/') + 1;
  sprintf(str,"FileName: %s", chp);
  cvPutText(BGRImg, str, cvPoint(1+10, 150), &font, CV_RGB(200,200,200));
  cvShowImage("Read", BGRImg);

  FrameNo++;
  sprintf(str,"FrameCount: %d", FrameCount);
  cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);
  cvPutText(BGRImg, str, cvPoint(1+10, 150), &font, CV_RGB(200,200,200));
  cvShowImage("Write", BGRImg);

  if(TempMode == 1){
    imageCopy(sfrom.x, sfrom.y, ReadBGRImg, SrchBGRImg);
    cvShowImage("SearchArea", SrchBGRImg);
  }

  switch(AutoDetect(SrchBGRImg, TempBGRImg)){
    case NOTFOUND:
      return;
    case OFFSETZERO:
      printf("Confirm the Position !\n");
  }

  while(1){
    #pragma omp parallel for
    for(i = 0; i < ImgSize; i++){
      accum = (unsigned int)AccImg[i];
      if(i - offset.x - Width * offset.y >= 0 && i - offset.x - Width * offset.y <= ImgSize){
        readi = (unsigned int)ReadImg[i - offset.x - Width * offset.y];
      }else{
        readi = 0;
      }
//      tmp = accum  + readi;
      tmp = accum  > readi ? accum  : readi;

      AccImg[i]   = (unsigned short)(tmp >  0xffff ?  0xffff : tmp);
      tmp = tmp >> 2;
      (BayerImg->imageData)[i] = (unsigned char)(tmp > 0xff ? 0xff : tmp);
    }

    cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);

    if(ZoomMode == 1){
      imageCopy(from.x, from.y, BGRImg, ZoomImg);
      cvShowImage("Zoom", ZoomImg);
      cvRectangle(BGRImg, cvPoint(from.x, from.y), cvPoint(to.x, to.y), CV_RGB(200,255,255), 2, 10, 0);
    }
    if(TempMode == 1)
      cvRectangle(BGRImg, cvPoint(sfrom.x, sfrom.y), cvPoint(sto.x, sto.y), CV_RGB(255,0,0), 2, 10, 0);
    cvShowImage("Work", BGRImg);

    int KeyCode = cvWaitKey(0);
    int SpecialCode = KeyCode / 0x10000;
    KeyCode &= 0x00ff;

    if(KeyCode == 'q' || KeyCode == ESC){
      stop_process = 1;
      break;
    }else if(KeyCode == 'y' || KeyCode == ' '){
      printf("%s Added\n", rfilename);
      #pragma omp parallel for
      for(i = 0; i < ImgSize; i++){
        WriteImg[i] = AccImg[i];
      }
      FrameCount++;
      break;
    }else{
      #pragma omp parallel for
      for(i = 0; i < ImgSize; i++){
        tmp = AccImg[i]   = WriteImg[i];
        tmp = tmp >> 2;
        (BayerImg->imageData)[i] = (unsigned char)(tmp > 0xff ? 0xff : tmp);
      }
      if(KeyCode == 'n'){
        printf("%s Not added\n", rfilename);
        break;
      }else if(KeyCode == 's'){
        SaveImage(FrameCount);
      }else if(KeyCode == 'a'){
        AutoDetect(SrchBGRImg, TempBGRImg);
      }else{
        int movement = 2;

        switch(SpecialCode){
          case ALT  : movement *= 5;
          case CTRL : movement *= 5;
          case SHIFT: movement *= 5;
          default:    break;
        }

        if(KeyCode == RIGHT     ) offset.x -= movement;
        else if(KeyCode == UP   ) offset.y -= movement;
        else if(KeyCode == LEFT ) offset.x += movement;
        else if(KeyCode == DOWN ) offset.y += movement;
      }
    }
  }
}


int main(int argc, char **argv){
    char globpath[256];
    char darkfpath[256];
    Width  = atoi(argv[1]);
    Height = atoi(argv[2]);
    char *dirname = argv[3];

    if(argc < 4){
      printf("SSP library still sample program in RAW16\n");
      printf("Copyright (c) Streamtechnology Corporation.\n");
      printf("              http://www.streamtechnology.co.jp/\n");
      printf("Usage: %s width height raw_file_dir_name [dark_filename]\n", argv[0]);

      return 0;
    }

    if(argc > 4){
      dark_filename = argv[4];
    }else{
      dark_filename = "dark.raw";
    }
    sprintf(globpath,"%s/*.raw",dirname);
    sprintf(darkfpath,"%s/%s",dirname,dark_filename);

    glob_t globbuf;
    int i;

    // OpenCV Settings
    cvNamedWindow("Read",  CV_WINDOW_KEEPRATIO); 
    cvNamedWindow("Write", CV_WINDOW_KEEPRATIO); 
    cvNamedWindow("Work",  CV_WINDOW_KEEPRATIO); 
    cvSetMouseCallback("Write", onMouse, "Write");
    cvSetMouseCallback("Work" , onMouse, "Work" );
    BayerImg     = cvCreateImage(cvSize(Width, Height),IPL_DEPTH_8U,1);
    ReadBayerImg = cvCreateImage(cvSize(Width, Height),IPL_DEPTH_8U,1);
    BGRImg       = cvCreateImage(cvSize(Width, Height),IPL_DEPTH_8U,3);
    ReadBGRImg   = cvCreateImage(cvSize(Width, Height),IPL_DEPTH_8U,3);
    cvInitFont(&font, CV_FONT_HERSHEY_SIMPLEX, 5.0, 5.0, 0, 10, CV_AA);

    AccImg   = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    ReadImg  = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    DarkImg  = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    WriteImg = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    #pragma omp parallel for
    for(i = 0; i < ImgSize; i++){
      AccImg[i]    = ReadImg[i]   = DarkImg[i]   = WriteImg[i]  = 0;
    }

    // Dark Frame file read
    FILE *file;
    file = fopen(darkfpath, "rb");
    if(file){
      fread(DarkImg, ImgSize, 2, file);
      fclose(file);
    }else{
      printf("Can't read Drak Frame File.\n");
    }

    // Image read for stack
    int ret = glob(globpath, 0, NULL, &globbuf);
    for (i = 0; i < globbuf.gl_pathc; i++) {
      if(strcmp(globbuf.gl_pathv[i],darkfpath) != 0){
        ReadImage(globbuf.gl_pathv[i]);
        if(stop_process == 1)
          break;
      }
    }
    globfree(&globbuf);

    if(FrameCount > 0 && stop_process == 0)
      SaveImage(FrameCount);

    free(AccImg  );
    free(ReadImg );
    free(DarkImg );
    free(WriteImg);
    cvReleaseImage(&BayerImg);
    cvReleaseImage(&ReadBayerImg);
    cvReleaseImage(&BGRImg);
    cvReleaseImage(&ReadBGRImg);
    cvReleaseImage(&ZoomImg);
    cvReleaseImage(&TempBGRImg);
    cvReleaseImage(&SrchBGRImg);
    cvDestroyWindow("Read");
    cvDestroyWindow("Write");
    cvDestroyWindow("Work");
    cvDestroyWindow("Zoom");
    cvDestroyWindow("Template");
    cvDestroyWindow("SearchArea");

    return 0;
}
