#include <stdio.h>
#include <string.h>
#include <glob.h>
#include <omp.h>
#include <opencv/cv.h>
#include <opencv/highgui.h>

// Key Code
#define SHIFT 0x01
#define CTRL  0x04
#define ALT   0x08
#define RIGHT 0x51
#define UP    0x52
#define LEFT  0x53
#define DOWN  0x54
#define ESC   0x1B

// 16bit raw Image Size
#define ImgSize Width*Height

int Width, Height;
int zWidth, zHeight;
int FrameNo    = 0;
int FrameCount = 0;
int ZoomMode   = 0;
IplImage *BGRImg, *BayerImg, *ZoomImg;
unsigned short *AccImg, *ReadImg, *DarkImg, *WriteImg;
char *dark_filename;
char str[100];
char wfilename[100];
char stop_process = 0;
CvFont font;
CvPoint from;
CvPoint to;

void imageCopy(int x, int y, int width, int height, IplImage *SrcImg, IplImage *DstImg){
  int i,j,k,l;
  for(j=from.y, l=0; l < height; j++, l++){
    for(i=from.x, k=0; k < width; i++, k++){
      (DstImg->imageData)[width * 3 * l + k * 3    ] = (SrcImg->imageData)[Width * 3 * j + i * 3    ];
      (DstImg->imageData)[width * 3 * l + k * 3 + 1] = (SrcImg->imageData)[Width * 3 * j + i * 3 + 1];
      (DstImg->imageData)[width * 3 * l + k * 3 + 2] = (SrcImg->imageData)[Width * 3 * j + i * 3 + 2];
    }
  }
}


void onMouse(int event, int x, int y, int flags, void* param){
  if(!strcmp(param, "Work")){
    if(event== CV_EVENT_LBUTTONDOWN){
      printf("From x:%03d y:%03d\n", x, y);
      from = cvPoint(x, y);
//    }else if(event== CV_EVENT_MOUSEMOVE && (flags & CV_EVENT_FLAG_LBUTTON)){
//      printf("To   x:%03d y:%03d\n", x, y);
//      to = cvPoint(x, y);
    }else if(event== CV_EVENT_LBUTTONUP){
      printf("To   x:%03d y:%03d\n", x, y);
      to = cvPoint(x, y);

      if(to.x - from.x < 0){
        to.x = from.x;
        from.x = x;
      }
      if(to.y - from.y < 0){
        to.y = from.y;
        from.y = y;
      }
      zWidth  = (to.x - from.x)/4*4;
      zHeight = (to.y - from.y)/4*4;
      to.x = from.x + zWidth;
      to.y = from.y + zHeight;
      if(zWidth == 0 || zHeight== 0)
        return;
      printf("Width:%d Height:%d\n",zWidth, zHeight);

      cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);
      ZoomImg = cvCreateImage(cvSize(zWidth, zHeight),IPL_DEPTH_8U,3);
      ZoomMode   = 1;
      imageCopy(from.x, from.y, zWidth, zHeight, BGRImg, ZoomImg);
      cvNamedWindow("Zoom", CV_WINDOW_KEEPRATIO); 
      cvShowImage("Zoom", ZoomImg);

      sprintf(str,"Frame No:%d", FrameNo);
      cvPutText(BGRImg, str, cvPoint(1+10, 150), &font, CV_RGB(200,200,200));
      cvRectangle(BGRImg, cvPoint(from.x, from.y), cvPoint(to.x, to.y), CV_RGB(200,255,255), 2, 10, 0);
      cvShowImage("Work", BGRImg);
    }
  }
}


void SaveImage(int FrameCount){
    sprintf(wfilename,"RawStackSL_%04d.raw", FrameCount);

    FILE *file;
    file = fopen(wfilename, "wb");
    if(file){
      fwrite(WriteImg, ImgSize, 2, file);
      fclose(file);
      printf("Saved RAW image to %s\n", wfilename);
    }
}


void ReadImage(char *rfilename){
  size_t i;
//  int offset_x = 0;
//  int offset_y = 0;
  unsigned int readi;
  unsigned int accum;
  unsigned int tmp;

  FILE *file;
  file = fopen(rfilename, "rb");
  if(file){
    fread(ReadImg, ImgSize, 2, file);
    fclose(file);
  }else{
    printf("Can't read Image File.\n");
  }

  #pragma omp parallel for
  for(i = 0; i < ImgSize; i++){
    unsigned int readi = (unsigned int)ReadImg[i];
    unsigned int darkv = (unsigned int)DarkImg[i];
    unsigned int tmp   = (readi > darkv) ? readi - darkv : 0;
    ReadImg[i]   = (unsigned short)tmp;
  }

  FrameNo++;
//  sprintf(str,"Frame Count:%d", FrameCount);
//  cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);
//  cvPutText(BGRImg, str, cvPoint(1+10, 150), &font, CV_RGB(200,200,200));
//  cvShowImage("Before", BGRImg);

    #pragma omp parallel for
    for(i = 0; i < ImgSize; i++){
      accum = (unsigned int)AccImg[i];
      readi = (unsigned int)ReadImg[i];
//      tmp = accum  + readi;
      tmp = accum  > readi ? accum  : readi;

      AccImg[i]   = (unsigned short)(tmp >  0xffff ?  0xffff : tmp);
      WriteImg[i] = AccImg[i];
//      tmp = tmp >> 2;
//      (BayerImg->imageData)[i] = (unsigned char)(tmp > 0xff ? 0xff : tmp);
    }

//    cvCvtColor(BayerImg, BGRImg, CV_BayerBG2BGR);
//    if(ZoomMode == 1){
//      imageCopy(from.x, from.y, zWidth, zHeight, BGRImg, ZoomImg);
//      cvShowImage("Zoom", ZoomImg);
//    }
//    sprintf(str,"Frame No:%d", FrameNo);
//    cvPutText(BGRImg, str, cvPoint(1+10, 150), &font, CV_RGB(200,200,200));
//    if(ZoomMode == 1){
//      cvRectangle(BGRImg, cvPoint(from.x, from.y), cvPoint(to.x, to.y), CV_RGB(200,255,255), 2, 10, 0);
//    }
//    cvShowImage("After", BGRImg);

    SaveImage(FrameCount);
    FrameCount++;
}


int main(int argc, char **argv){
    char globpath[256];
    char darkfpath[256];
    Width  = atoi(argv[1]);
    Height = atoi(argv[2]);
    char *dirname = argv[3];

    if(argc < 4){
      printf("SSP library still sample program in RAW16\n");
      printf("Copyright (c) Streamtechnology Corporation.\n");
      printf("              http://www.streamtechnology.co.jp/\n");
      printf("Usage: %s width height raw_file_dir_name [dark_filename]\n", argv[0]);

      return 0;
    }

    if(argc > 4){
      dark_filename = argv[4];
    }else{
      dark_filename = "dark.raw";
    }
    sprintf(globpath,"%s/*.raw",dirname);
    sprintf(darkfpath,"%s/%s",dirname,dark_filename);

    glob_t globbuf;
    int i;

    // OpenCV Settings
    cvNamedWindow("Write", CV_WINDOW_KEEPRATIO); 
    cvNamedWindow("Work",  CV_WINDOW_KEEPRATIO); 
    cvSetMouseCallback("Write", onMouse, "Write");
    cvSetMouseCallback("Work" , onMouse, "Work" );
    BayerImg = cvCreateImage(cvSize(Width, Height),IPL_DEPTH_8U,1);
    BGRImg   = cvCreateImage(cvSize(Width, Height),IPL_DEPTH_8U,3);
    cvInitFont(&font, CV_FONT_HERSHEY_SIMPLEX, 5.0, 5.0, 0, 10, CV_AA);

    AccImg   = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    ReadImg  = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    DarkImg  = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    WriteImg = (unsigned short *) malloc(sizeof(unsigned short) * ImgSize);
    #pragma omp parallel for
    for(i = 0; i < ImgSize; i++){
      AccImg[i]    = ReadImg[i]   = DarkImg[i]   = WriteImg[i]  = 0;
    }

    // Dark Frame file read
    FILE *file;
    file = fopen(darkfpath, "rb");
    if(file){
      fread(DarkImg, ImgSize, 2, file);
      fclose(file);
    }else{
      printf("Can't read Drak Frame File.\n");
    }

    // Image read for stack
    int ret = glob(globpath, 0, NULL, &globbuf);
    for (i = 0; i < globbuf.gl_pathc; i++) {
      if(strcmp(globbuf.gl_pathv[i],darkfpath) != 0){
        ReadImage(globbuf.gl_pathv[i]);
        if(stop_process == 1)
          break;
      }
    }
    globfree(&globbuf);

    if(FrameCount > 0 && stop_process == 0)
      SaveImage(FrameCount);

    free(AccImg  );
    free(ReadImg );
    free(DarkImg );
    free(WriteImg);
    cvReleaseImage(&BayerImg);
    cvReleaseImage(&BGRImg);
    cvDestroyWindow("Write");
    cvDestroyWindow("Work");

    return 0;
}
