#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <libssp.h>
#include <libsspprof.h>
#include <wiringPi.h>
#include <signal.h>
#include<unistd.h>
#include <sys/time.h>
#include "../common/isreg.h"


#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE

int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

struct isreg *isregp;
static int frame_count;

char *filename;
int stop_process;
static void signal_handler(int sig)
{
  printf("Terminated.\n");
  if(sig == SIGINT)
    stop_process = 1;
}


void frame_drop_func(struct ssp_handle *handle){

  printf("Frame dropped by camera.\n");

}

void frame_drop_preprocess_func(struct ssp_handle *handle){

  printf("Frame dropped at preprocess.\n");
}

//#define WaitCount 1000
#define WaitCount 5
void fame_preprocess(struct ssp_handle *handle, struct ssp_frame *frame){
  //  printf("Preprocess is active. frame_count = %d\n", frame_count);
  frame_count ++;
  if(frame_count % 10 == 0) printf("frame_count = %d\n", frame_count);
  usleep(1000);
  if(frame_count == WaitCount){
    FILE *file;
    file = fopen(filename, "wb");
    if(file){
      size_t size = 0;
      if(ssp_get_frame_size(frame, &size) != SSP_SUCCESS){
	printf("Unkown frame type.\n");
	ssp_release_frame(frame);
	return;
      }
      fwrite(ssp_get_frame_data(frame), size, 1, file);
      fclose(file);
      printf("Saved RAW image to %s\n", filename);
    }
  }

  ssp_release_frame(frame);

}


int main(int argc, char **argv){
  if(argc < 6){
    printf("SSP library still sample program in RAW16\n");
    printf("Copyright (c) Streamtechnology Corporation.\n");
    printf("              http://www.streamtechnology.co.jp/\n");
    printf("Usage: %s profile_file_name model_name output_file_name Gain Exposure\n", argv[0]);

    return 0;
  }

  char *profile_filename;
  char *model_name;
  int RegAdr_Gain;
  int RegAdr_ExpL;
  int RegAdr_ExpH;
  int RegVal_Gain;
  int RegVal_ExpL;
  int RegVal_ExpH;

  profile_filename = argv[1];
  model_name = argv[2];
  filename = argv[3];
  
  RegVal_Gain = atoi(argv[4]);
  RegVal_ExpL = atoi(argv[5]) & 0xFF;
  RegVal_ExpH = atoi(argv[5]) / 0x0100;

  isregp = (struct isreg *)check_name(model_name);
  if(isregp == NULL){
    printf("not found name %s\n",model_name);
    return(1);
  }

  stop_process = 0;
  if (signal(SIGINT, signal_handler) == SIG_ERR)
    return 2;
  
  struct ssp_settings settings;

  frame_count = 0;
  
  settings.lib_settings.NumFrameFIFOSize = 16;
  settings.lib_settings.NumThreadForBuildInPreprocess = 1;
  settings.lib_settings.frame_drop_cam_user_func = frame_drop_func;
  settings.lib_settings.frame_drop_pre_user_func = frame_drop_preprocess_func;
  settings.lib_settings.fame_preprocess_user_func = fame_preprocess;
  settings.frame_preprocess_options.FormatConvertType = SSP_FRAME_BAYER16;
  settings.frame_preprocess_options.GammaCorrectionEnable = SSP_DISABLE;
  settings.frame_preprocess_options.GammaVal = 2.2;
  settings.camera_settings.PowerOnResetUsec = 1000000;

  struct ssp_handle *handle;

  struct ssp_profile *profile;
  
  ssp_read_profile(profile_filename, &profile);

  printf("Setting summary: %d x %d @ %d fps %d bits\n",
	 profile->ImageProperty.Width,
	 profile->ImageProperty.Height,
	 profile->ImageProperty.Frequency,
	 profile->ImageProperty.BayerBits);

  int i,j;
  for(i=0;i<1;i++){ // for repeat initialization test.
    int retval;
    if((retval = ssp_initialize(&handle, profile, &settings))!=SSP_SUCCESS){
      
      printf("Can not initialize ssp library. %x\n",retval);
      return 0;
    }

    //Analog Gain set
    __CCIRegWrite(handle, isregp->Adr_GainL, RegVal_Gain & 0xFF);
    if(isregp->Adr_GainH != -1){
      __CCIRegWrite(handle, isregp->Adr_GainH, (RegVal_Gain>>8)&0x1);
    }
    __CCIRegWrite(handle, isregp->Adr_ExpH, RegVal_ExpH);
    __CCIRegWrite(handle, isregp->Adr_ExpL, RegVal_ExpL);
    printf("Gain value = %d\n",RegVal_Gain);
    printf("Initialize done.\n");
    printf("start streaming\n");
    if((retval = ssp_start_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }

    while(frame_count < WaitCount){
      if(stop_process != 0) break;
    }
    printf("FrameCount = %d\n", frame_count);
    printf("stop streaming\n");
    if((retval = ssp_stop_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }
    
    printf("Sync Wait.\n");
    ssp_sync_event(handle);
    printf("Sync Done.\n");

    if(ssp_finalize(handle) != SSP_SUCCESS){
      printf("Error during finalization.\n");
      return 0;
    }
    printf("Finalize done.\n");fflush(stdout);
  }
  return 1;
}
