
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <libssp.h>
#include <wiringPi.h>
#include <signal.h>

#include <omp.h>
#include<opencv/cv.h>
#include <opencv/highgui.h>
// sudo apt-get install libopencv-dev

#include <sys/time.h>
#include "../common/isreg.h"


#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE

int ProcKey(int KeyCode,struct ssp_handle *handle);
int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

struct isreg *isregp;
static int frame_count;
static int Gain;
volatile static int coarse_integration_time;
int magnification = 1;

IplImage *RGBImg, *BayerImg;

char *filename;
volatile int stop_process;

static void signal_handler(int sig)
{
  printf("Terminated.\n");
  if(sig == SIGINT)
    stop_process = 1;
}


void frame_drop_func(struct ssp_handle *handle){

    printf("Frame dropped by camera.\n");
    stop_process = 1;

}

void frame_drop_preprocess_func(struct ssp_handle *handle){

  printf("Frame dropped at preprocess.\n");
}

CvVideoWriter *VideoWriter;

//#define WaitMaxCount 200
int WaitMaxCount;
//struct ssp_frame **frame_buffer;

#define EXPOSE 0
#define PAUSE 1
#define AVIWRITE 2
int ExpStatus = PAUSE;
int GammaEnable = 0;
char str[100];
CvFont font;

void GammaCorrection(IplImage* src, IplImage* dst, double gamma){
  int i;
  uchar LUT[256];

  for (i = 0; i < 256; i++){
    LUT[i] = (int)(pow((double)i / 255.0, 1.0 / gamma) * 255.0);
  }
  CvMat lut_mat = cvMat(1, 256, CV_8UC1, LUT);
  cvLUT(src, dst, &lut_mat);
}

struct timeval tv_start, tv_end;
struct timezone tz_local;

void fame_preprocess(struct ssp_handle *handle, struct ssp_frame *frame){
  char filenameAVI[128];
  static int filecount = 0;
  size_t i;
  int j;
  size_t size = 0;
  
  unsigned char *frame_data = ssp_get_frame_data(frame);
  #pragma omp parallel for
  for(i=0;i<BayerImg->height;i++)
    #pragma omp parallel for
    for(j=0;j<BayerImg->width;j++){
      ((unsigned char *)(BayerImg->imageData + BayerImg->widthStep * i))[j] = frame_data[i*BayerImg->widthStep+j];
    }
  ssp_release_frame(frame);
  cvCvtColor( BayerImg, RGBImg, CV_BayerRG2RGB);

  if(GammaEnable == 1)
    GammaCorrection(RGBImg, RGBImg, 2.2);

  if(ExpStatus == EXPOSE){
    frame_count ++;

    if(frame_count == 1){
      filecount++;
      double fps = (double)magnification;
      printf("FrameRate = %f\n", fps);
      printf("FrameCount = %d\n", frame_count);
      sprintf(filenameAVI,"%s_%04d_g%d_c%d.avi", filename, filecount, Gain, coarse_integration_time);
      VideoWriter = cvCreateVideoWriter(filenameAVI,CV_FOURCC('M','J','P','G'),
//      VideoWriter = cvCreateVideoWriter(filenameAVI,0,
                                   fps,
                                   cvSize(ssp_get_frame_width(frame),
                                   ssp_get_frame_height(frame)), 1);
    }

    printf("frame_count = %d\n", frame_count);

    if(VideoWriter != NULL){
      cvWriteFrame(VideoWriter, RGBImg);
    }else{
      printf("OpenCV Error: VideoWriter\n");
    }

    if(frame_count >= WaitMaxCount){
      cvReleaseVideoWriter(&VideoWriter);
      ExpStatus = PAUSE;
      frame_count  = 0;
      stop_process = 0;
      printf("PAUSE\n");
    }
  }

  sprintf(str,"Gain:%d Expose:%d", Gain, coarse_integration_time);
  cvPutText(RGBImg, str, cvPoint(1+10*ssp_get_frame_width(frame)/3200, 150*ssp_get_frame_height(frame)/2400), &font, CV_RGB(200,200,200));
  cvShowImage("SSP Viewer", RGBImg);

  ProcKey(cvWaitKey(2) & 0xFF,handle);
}

char *model_name;
#define GAIN_CHANGE_VAL 25
#define GAIN_MODE 0
#define SS_MODE 1 // Shutter Speed
int SettingMode = 0;

int ProcKey(int KeyCode,struct ssp_handle *handle){

  if(KeyCode == 'q'){
    stop_process = 1;
    printf("Quit\n");
  }else if(KeyCode == ' '){
    ExpStatus = EXPOSE;
    printf("Exposure\n");
  }else if(KeyCode == 's'){
    printf("Setting Mode is set to Shutter Speed Mode.\n");
    SettingMode = SS_MODE;
  }else if(KeyCode == 'g'){
    printf("Setting Mode is set to Gain Mode.\n");
    SettingMode = GAIN_MODE;
  }else if(KeyCode == '0'){
    if(GammaEnable == 0){
      GammaEnable = 1;
      printf("Gamma Correction is Enabled.\n");
    }
    else{
      GammaEnable = 0;
      printf("Gamma Correction is Disabled.\n");
    }
  }

  if(KeyCode == 'y' || KeyCode == 'h' || KeyCode == 'u' || KeyCode == 'j' || KeyCode == 'i' || KeyCode == 'k' || KeyCode == 'o' || KeyCode == 'l' || KeyCode == 'p' || KeyCode == ';'){
    int RegVal;
    int step = 0;

    if(KeyCode == 'y')      step =  1;
    else if(KeyCode == 'u') step =  10;
    else if(KeyCode == 'i') step =  100;
    else if(KeyCode == 'o') step =  1000;
    else if(KeyCode == 'p') step =  10000;
    else if(KeyCode == 'h') step = -1;
    else if(KeyCode == 'j') step = -10;
    else if(KeyCode == 'k') step = -100;
    else if(KeyCode == 'l') step = -1000;
    else if(KeyCode == ';') step = -10000;

    if(SettingMode == GAIN_MODE){
      int GainVal;
      if(__CCIRegRead(handle, isregp->Adr_GainL, &RegVal) == 0)
        printf("Reading register error.\n");
      GainVal = RegVal;
      if(isregp->Adr_GainH != -1){
        if(__CCIRegRead(handle, isregp->Adr_GainH, &RegVal) == 0)
          printf("Reading register error.\n");
        GainVal |= ((RegVal & 0x1) << 8);
      }
      RegVal = GainVal;

      RegVal += step;
      if(RegVal > isregp->Gain_Max) RegVal = isregp->Gain_Min;
      else if(RegVal < isregp->Gain_Min) RegVal = isregp->Gain_Max;

      if(__CCIRegWrite(handle, isregp->Adr_GainL, RegVal & 0xFF) == 0)
        printf("Writing register error.\n");
      if(isregp->Adr_GainH != -1){
        if(__CCIRegWrite(handle, isregp->Adr_GainH, (RegVal>>8)&0x1) == 0)
          printf("Writing register error.\n");        
      }
      if(__CCIRegRead(handle, isregp->Adr_GainL, &RegVal) == 0)
        printf("Reading register error.\n");
      GainVal = RegVal;
      if(isregp->Adr_GainH != -1){
        if(__CCIRegRead(handle, isregp->Adr_GainH, &RegVal) == 0)
          printf("Reading register error.\n");
        GainVal |= ((RegVal & 0x1) << 8);
      }
      RegVal = GainVal;
      printf("Analog Gain Set to %d\n", RegVal);
      Gain = RegVal;
    }

    if(SettingMode == SS_MODE){
      int frame_length_lines;
      
      if(__CCIRegRead(handle, isregp->Adr_FllH, &RegVal) == 0)
        printf("Reading register error.\n");
      frame_length_lines = RegVal<<8;
      if(__CCIRegRead(handle, isregp->Adr_FllL, &RegVal) == 0)
        printf("Reading register error.\n");
      frame_length_lines |= RegVal;

      if(__CCIRegRead(handle, isregp->Adr_ExpH, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time = RegVal<<8;
      if(__CCIRegRead(handle, isregp->Adr_ExpL, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time |= RegVal;

      coarse_integration_time += step;
      if(coarse_integration_time >= (frame_length_lines-5)) RegVal = 1;
      else if(coarse_integration_time <= -1) coarse_integration_time = frame_length_lines-4;

      if(__CCIRegWrite(handle, isregp->Adr_ExpH, (coarse_integration_time>>8)&0xFF) == 0)
        printf("Writing register error.\n");
      if(__CCIRegWrite(handle, isregp->Adr_ExpL, coarse_integration_time&0xFF) == 0)
        printf("Writing register error.\n");
      if(__CCIRegRead(handle, isregp->Adr_ExpH, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time = RegVal<<8;
      if(__CCIRegRead(handle, isregp->Adr_ExpL, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time |= RegVal;
      printf("Shutter Speed Set (coarse_integration_time) to %d\n", coarse_integration_time);
    }
  }
  return 0;
}



int main(int argc, char **argv){

  if(argc < 5){
    printf("SSP library video sample program in Motion JPEG\n");
    printf("Copyright (c) Streamtechnology Corporation.\n");
    printf("              http://www.streamtechnology.co.jp/\n");
    printf("Usage: %s profile_file model_name output_file_name WaitMaxCount Gain Exposure fps\n", argv[0]);
    return 0;
  }

  char *profile_filename;
  profile_filename = argv[1];
  model_name = argv[2];
  filename = argv[3];

  WaitMaxCount = atoi(argv[4]);

  int RegVal_Gain = atoi(argv[5]);
  int RegVal_ExpL = atoi(argv[6]) & 0xFF;
  int RegVal_ExpH = atoi(argv[6]) / 0x0100;
  Gain = RegVal_Gain;
  coarse_integration_time = atoi(argv[6]);
  if(argc = 6)
    magnification = atoi(argv[7]);

  isregp = (struct isreg *)check_name(model_name);
  if(isregp == NULL){
    printf("not found name %s\n",model_name);
    return(1);
  }

  stop_process = 0;
  if (signal(SIGINT, signal_handler) == SIG_ERR)
    return 2;

  struct ssp_settings settings;

  frame_count = 0;
  
  // SSP settings
  settings.lib_settings.NumFrameFIFOSize = 2;
  settings.lib_settings.NumThreadForBuildInPreprocess = 1;
  settings.lib_settings.frame_drop_cam_user_func = frame_drop_func;
  settings.lib_settings.frame_drop_pre_user_func = frame_drop_preprocess_func;
  settings.lib_settings.fame_preprocess_user_func = fame_preprocess;
  settings.frame_preprocess_options.FormatConvertType = SSP_FRAME_BAYER8;
  settings.frame_preprocess_options.GammaCorrectionEnable = GAMMA_OPTION;
  settings.frame_preprocess_options.GammaVal = 2.2;
  settings.camera_settings.PowerOnResetUsec = 1000000;

  struct ssp_handle *handle;

  struct ssp_profile *profile;
  
  ssp_read_profile(profile_filename, &profile);

  printf("Setting summary: %d x %d @ %d fps %d bits\n",
	 profile->ImageProperty.Width,
	 profile->ImageProperty.Height,
	 profile->ImageProperty.Frequency,
	 profile->ImageProperty.BayerBits);

  // OpenCV Settings
  cvNamedWindow("SSP Viewer", CV_WINDOW_KEEPRATIO); 
  BayerImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,1);
  RGBImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,3);
  cvInitFont(&font, CV_FONT_HERSHEY_SIMPLEX, 5.0*profile->ImageProperty.Width/3200, 5.0*profile->ImageProperty.Height/2400, 0, 10*profile->ImageProperty.Width/3200, CV_AA);

  int i,j;
  for(i=0;i<1;i++){ // for repeat initialization test.
    int retval;
    if((retval = ssp_initialize(&handle, profile, &settings))!=SSP_SUCCESS){
      
      printf("Can not initialize ssp library. %x\n",retval);
      return 0;
    }

    //Analog Gain set
    __CCIRegWrite(handle, isregp->Adr_GainL, RegVal_Gain & 0xFF);
    if(isregp->Adr_GainH != -1){
      __CCIRegWrite(handle, isregp->Adr_GainH, (RegVal_Gain>>8)&0x1);
    }
    __CCIRegWrite(handle, isregp->Adr_ExpH, RegVal_ExpH);
    __CCIRegWrite(handle, isregp->Adr_ExpL, RegVal_ExpL);
    printf("Gain value = %d\n",RegVal_Gain);
    printf("coarse_integration_time = %d\n",coarse_integration_time);
    printf("Initialize done.\n");
    printf("start streaming\n");
    if((retval = ssp_start_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }

    
    while(1){
      if(stop_process != 0) break;
    }

    printf("Flush Wait.\n");
    ssp_flush_event(handle);
    printf("Flush Done.\n");

    if(ssp_finalize(handle) != SSP_SUCCESS){
      printf("Error during finalization.\n");
//      return 0;
    }

//    cvReleaseVideoWriter(&VideoWriter);
    cvReleaseImage(&RGBImg);
    cvReleaseImage(&BayerImg);
    cvDestroyWindow("SSP Viewer");

    printf("Finalize done.\n");fflush(stdout);
  }
  return 1;
}
