#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <libssp.h>
#include <wiringPi.h>
#include <signal.h>
#include <unistd.h>
#include <opencv/cv.h>
#include <opencv/highgui.h>
// sudo apt-get install libopencv-dev
// sudo apt-get install libgl1-mesa-dri

#include <sys/time.h>
#include <pthread.h>
#include "../common/isreg.h"

#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE

int ProcKey(int KeyCode,struct ssp_handle *handle);
int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

struct isreg *isregp;
static int frame_count;
static int Gain;
static int coarse_integration_time;

IplImage* BayerImg, *RGBImg;
IplImage *ZoomImg;

volatile int stop_process;

#define ZOOM_INIT   0
#define ZOOM_SELECT 1
#define ZOOM_ON     2

int zWidth, zHeight;
int ZoomMode   = ZOOM_INIT;
CvPoint from;
CvPoint to;
CvPoint mfrom;
CvPoint mto;

void imageCopy(int x, int y, IplImage *SrcImg, IplImage *DstImg){
  int i,j,k,l;
  for(j=y, l=0; l < DstImg->height; j++, l++){
    for(i=x, k=0; k < DstImg->width; i++, k++){
      (DstImg->imageData)[DstImg->width * 3 * l + k * 3    ] = (SrcImg->imageData)[SrcImg->width * 3 * j + i * 3    ];
      (DstImg->imageData)[DstImg->width * 3 * l + k * 3 + 1] = (SrcImg->imageData)[SrcImg->width * 3 * j + i * 3 + 1];
      (DstImg->imageData)[DstImg->width * 3 * l + k * 3 + 2] = (SrcImg->imageData)[SrcImg->width * 3 * j + i * 3 + 2];
    }
  }
}


void onMouse(int event, int x, int y, int flags, void* param){
  if(!strcmp(param, "SSP Viewer")){
    if(event== CV_EVENT_LBUTTONDOWN){
      printf("From x:%03d y:%03d\n", x, y);
      mfrom = cvPoint(x, y);
      mto = cvPoint(x, y);
      ZoomMode = ZOOM_SELECT;
    }else if(event== CV_EVENT_MOUSEMOVE && ZoomMode == ZOOM_SELECT){
      printf("To   x:%03d y:%03d\n", x, y);
      mto = cvPoint(x, y);
      if((mto.x - mfrom.x)/4*4 == 0 || (mto.y - mfrom.y)/4*4 == 0)
        return;

      if(mto.x < 0 || mto.x > RGBImg->width){
      }else if(mto.x < mfrom.x){
        to.x   = mfrom.x;
        from.x = mto.x;
      }else{
        from.x = mfrom.x;
        to.x   = mto.x;
      }
      if(mto.y < 0 || mto.y > RGBImg->height){
      }else if(mto.y < mfrom.y){
        to.y   = mfrom.y;
        from.y = mto.y;
      }else{
        from.y = mfrom.y;
        to.y   = mto.y;
      }

      zWidth  = (to.x - from.x)/4*4;
      zHeight = (to.y - from.y)/4*4;
      to.x = from.x + zWidth;
      to.y = from.y + zHeight;
      if(zWidth == 0 || zHeight== 0)
        return;
    }else if(event== CV_EVENT_LBUTTONUP){
      printf("Width:%d Height:%d\n",zWidth, zHeight);
      cvNamedWindow("Zoom", CV_WINDOW_AUTOSIZE); 
      ZoomImg = cvCreateImage(cvSize(zWidth, zHeight),IPL_DEPTH_8U,3);
      ZoomMode   = ZOOM_ON;
    }
  }
}


static void signal_handler(int sig)
{
  printf("Terminated.\n");
  if(sig == SIGINT)
    stop_process = 1;
}


void frame_drop_func(struct ssp_handle *handle){

    printf("Frame dropped by camera.\n");

}

void frame_drop_preprocess_func(struct ssp_handle *handle){

  printf("Frame dropped at preprocess.\n");
}


void GammaCorrection(IplImage* src, IplImage* dst, double gamma){
  int i;
  uchar LUT[256];

  for (i = 0; i < 256; i++){
    LUT[i] = (int)(pow((double)i / 255.0, 1.0 / gamma) * 255.0);
  }
  CvMat lut_mat = cvMat(1, 256, CV_8UC1, LUT);
  cvLUT(src, dst, &lut_mat);
}

volatile int GammaEnable = 0;

struct __paracopy_args {
  int width;
  int height;
  unsigned char *src;
  unsigned char *dest;
};

void *paracopy_thread_func(void *param){
  struct __paracopy_args *copyparams = (struct __paracopy_args *)param;
  int i;
  for(i=0;i<copyparams->height;i++)
    memcpy(copyparams->dest+copyparams->width*i,
           copyparams->src+copyparams->width*i,
           copyparams->width);

  return NULL;
}


#define NUM_COPY_THREAD 4
pthread_t thread_handle[NUM_COPY_THREAD];
struct __paracopy_args copy_params[NUM_COPY_THREAD];
char str[100];
CvFont font;


void fame_preprocess(struct ssp_handle *handle, struct ssp_frame *frame){
  int i;
  unsigned char *frame_data = ssp_get_frame_data(frame);
#if 0 // Simple copy
  for(i=0;i<BayerImg->height;i++)
    for(j=0;j<BayerImg->width;j++){
      ((unsigned char *)(BayerImg->imageData + BayerImg->widthStep * i))[j] = frame_data[i*BayerImg->widthStep+j];
    }
#endif

#if 0 // memcpy
  for(i=0;i<BayerImg->height;i++){
    memcpy((unsigned char *)(BayerImg->imageData + BayerImg->widthStep * i), &frame_data[i*BayerImg->widthStep], BayerImg->width);
  }
#endif
#if 0 // long long copy
  for(i=0;i<BayerImg->height;i++)
    for(j=0;j<BayerImg->width/sizeof(long long);j++){
      *((long long *)(BayerImg->imageData + BayerImg->widthStep * i) + j) = *((long long*)(frame_data+i*BayerImg->widthStep)+j);
    }
#endif

#if 1 // parallel copy
  for(i=0;i<NUM_COPY_THREAD;i++){
    copy_params[i].width = BayerImg->widthStep;
    copy_params[i].height = BayerImg->height/NUM_COPY_THREAD;
    copy_params[i].dest = (unsigned char *)(BayerImg->imageData + BayerImg->widthStep * (i*BayerImg->height/NUM_COPY_THREAD));
    copy_params[i].src = &frame_data[(i*BayerImg->height/NUM_COPY_THREAD)*BayerImg->widthStep];
    pthread_create(&(thread_handle[i]) , NULL , paracopy_thread_func ,&(copy_params[i]));
      
  }
  for(i=0;i<NUM_COPY_THREAD;i++)
    pthread_join(thread_handle[i] , NULL);
#endif
  
  ssp_release_frame(frame);

  cvCvtColor( BayerImg, RGBImg, CV_BayerRG2RGB);
  if(GammaEnable == 1)
    GammaCorrection(RGBImg, RGBImg, 2.2);

  if(ZoomMode == ZOOM_ON)
    imageCopy(from.x, from.y, RGBImg, ZoomImg);

  if(ZoomMode != ZOOM_INIT){
    cvShowImage("Zoom", ZoomImg);
    cvRectangle(RGBImg, cvPoint(from.x, from.y), cvPoint(to.x, to.y), CV_RGB(0,0,0), 12, 10, 0);
    cvRectangle(RGBImg, cvPoint(from.x, from.y), cvPoint(to.x, to.y), CV_RGB(255,0,0), 4, 10, 0);
  }

  sprintf(str,"Gain:%d Expose:%d", Gain, coarse_integration_time);
  cvPutText(RGBImg, str, cvPoint(1+10*ssp_get_frame_width(frame)/3200, 150*ssp_get_frame_height(frame)/2400), &font, CV_RGB(200,200,200));
  cvShowImage("SSP Viewer", RGBImg);

  ProcKey(cvWaitKey(2) & 0xFF,handle);

  printf("frame  count = %d\n", frame_count);
  frame_count ++;
}


char *model_name;
#define GAIN_MODE 0
#define SS_MODE   1 // Shutter Speed
int SettingMode = 0;

int ProcKey(int KeyCode,struct ssp_handle *handle){
  if(KeyCode == 'q'){
    stop_process = 1;
    printf("Quit\n");
  }else if(KeyCode == 's'){
    printf("Setting Mode is set to Shutter Speed Mode.\n");
    SettingMode = SS_MODE;
  }else if(KeyCode == 'g'){
    printf("Setting Mode is set to Gain Mode.\n");
    SettingMode = GAIN_MODE;
  }else if(KeyCode == '0'){
    if(GammaEnable == 0){
      GammaEnable = 1;
      printf("Gamma Correction is Enabled.\n");
    }
    else{
      GammaEnable = 0;
      printf("Gamma Correction is Disabled.\n");
    }
  }
  if(KeyCode == 'y' || KeyCode == 'h' || KeyCode == 'u' || KeyCode == 'j' || KeyCode == 'i' || KeyCode == 'k' || KeyCode == 'o' || KeyCode == 'l' || KeyCode == 'p' || KeyCode == ';'){
    int RegVal;
    int step = 0;

    if(KeyCode == 'y')      step =  1;
    else if(KeyCode == 'u') step =  10;
    else if(KeyCode == 'i') step =  100;
    else if(KeyCode == 'o') step =  1000;
    else if(KeyCode == 'p') step =  10000;
    else if(KeyCode == 'h') step = -1;
    else if(KeyCode == 'j') step = -10;
    else if(KeyCode == 'k') step = -100;
    else if(KeyCode == 'l') step = -1000;
    else if(KeyCode == ';') step = -10000;

    if(SettingMode == GAIN_MODE){
      int GainVal;
      if(__CCIRegRead(handle, isregp->Adr_GainL, &RegVal) == 0)
        printf("Reading register error.\n");
      GainVal = RegVal;
      if(isregp->Adr_GainH != -1){
        if(__CCIRegRead(handle, isregp->Adr_GainH, &RegVal) == 0)
          printf("Reading register error.\n");
        GainVal |= ((RegVal & 0x1) << 8);
      }
      RegVal = GainVal;

      RegVal += step;
      if(RegVal > isregp->Gain_Max) RegVal = isregp->Gain_Min;
      else if(RegVal < isregp->Gain_Min) RegVal = isregp->Gain_Max;

      if(__CCIRegWrite(handle, isregp->Adr_GainL, RegVal & 0xFF) == 0)
        printf("Writing register error.\n");
      if(isregp->Adr_GainH != -1){
        if(__CCIRegWrite(handle, isregp->Adr_GainH, (RegVal>>8)&0x1) == 0)
          printf("Writing register error.\n");        
      }
      if(__CCIRegRead(handle, isregp->Adr_GainL, &RegVal) == 0)
        printf("Reading register error.\n");
      GainVal = RegVal;
      if(isregp->Adr_GainH != -1){
        if(__CCIRegRead(handle, isregp->Adr_GainH, &RegVal) == 0)
          printf("Reading register error.\n");
        GainVal |= ((RegVal & 0x1) << 8);
      }
      RegVal = GainVal;
      printf("Analog Gain Set to %d\n", RegVal);
      Gain = RegVal;
    }

    if(SettingMode == SS_MODE){
      int frame_length_lines;
      
      if(__CCIRegRead(handle, isregp->Adr_FllH, &RegVal) == 0)
        printf("Reading register error.\n");
      frame_length_lines = RegVal<<8;
      if(__CCIRegRead(handle, isregp->Adr_FllL, &RegVal) == 0)
        printf("Reading register error.\n");
      frame_length_lines |= RegVal;

      if(__CCIRegRead(handle, isregp->Adr_ExpH, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time = RegVal<<8;
      if(__CCIRegRead(handle, isregp->Adr_ExpL, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time |= RegVal;

      coarse_integration_time += step;
      if(coarse_integration_time >= (frame_length_lines-5)) RegVal = 1;
      else if(coarse_integration_time <= -1) coarse_integration_time = frame_length_lines-4;

      if(__CCIRegWrite(handle, isregp->Adr_ExpH, (coarse_integration_time>>8)&0xFF) == 0)
        printf("Writing register error.\n");
      if(__CCIRegWrite(handle, isregp->Adr_ExpL, coarse_integration_time&0xFF) == 0)
        printf("Writing register error.\n");
      if(__CCIRegRead(handle, isregp->Adr_ExpH, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time = RegVal<<8;
      if(__CCIRegRead(handle, isregp->Adr_ExpL, &RegVal) == 0)
        printf("Reading register error.\n");
      coarse_integration_time |= RegVal;
      printf("Shutter Speed Set (coarse_integration_time) to %d\n", coarse_integration_time);
    }
  }
  return 0;
}


int main(int argc, char **argv){
  int RegVal_Gain;
  int RegVal_ExpL;
  int RegVal_ExpH;

  if(argc < 5){
    printf("SSP library viewer sample program\n");
    printf("Copyright (c) Streamtechnology Corporation.\n");
    printf("              http://www.streamtechnology.co.jp/\n");
    printf("Usage: %s profile_file model Gain Exposure\n", argv[0]);
    printf("\tmodel: IMX219 or IMX408 or IU233 or IMX378\n");
    return 0;
  }

  char *profile_filename;
  profile_filename = argv[1];
  model_name = argv[2];
  RegVal_Gain = atoi(argv[3]);
  RegVal_ExpL = atoi(argv[4]) & 0xFF;
  RegVal_ExpH = atoi(argv[4]) / 0x0100;
  Gain = RegVal_Gain;
  coarse_integration_time = atoi(argv[4]);

  isregp = (struct isreg *)check_name(model_name);
  if(isregp == NULL){
    printf("not found name %s\n",model_name);
    return(1);
  }

  stop_process = 0;
  if (signal(SIGINT, signal_handler) == SIG_ERR)
    return 2;

  struct ssp_settings settings;

  frame_count = 0;
  
  settings.lib_settings.NumFrameFIFOSize = 2;
  settings.lib_settings.NumThreadForBuildInPreprocess = 4;
  settings.lib_settings.frame_drop_cam_user_func = frame_drop_func;
  settings.lib_settings.frame_drop_pre_user_func = frame_drop_preprocess_func;
  settings.lib_settings.fame_preprocess_user_func = fame_preprocess;
  settings.frame_preprocess_options.FormatConvertType = SSP_FRAME_BAYER8;
  settings.frame_preprocess_options.GammaCorrectionEnable = GAMMA_OPTION;
  settings.frame_preprocess_options.GammaVal = 2.2;
  settings.camera_settings.PowerOnResetUsec = 1000000;

  struct ssp_handle *handle;

  struct ssp_profile *profile;
  
  ssp_read_profile(profile_filename, &profile);

  printf("Setting summary: %d x %d @ %d fps %d bits\n",
         profile->ImageProperty.Width,
         profile->ImageProperty.Height,
         profile->ImageProperty.Frequency,
         profile->ImageProperty.BayerBits);


  // OpenCV Settings
  cvNamedWindow("SSP Viewer", CV_WINDOW_KEEPRATIO); 
  cvSetMouseCallback("SSP Viewer" , onMouse, "SSP Viewer" );

  BayerImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,1);
  RGBImg = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,3);
  cvInitFont(&font, CV_FONT_HERSHEY_SIMPLEX, 5.0*profile->ImageProperty.Width/3200, 5.0*profile->ImageProperty.Height/2400, 0, 10*profile->ImageProperty.Width/3200, CV_AA);

  
  int i;
  for(i=0;i<1;i++){ // for repeat initialization test.
    int retval;
    if((retval = ssp_initialize(&handle, profile, &settings))!=SSP_SUCCESS){
      
      printf("Can not initialize ssp library. %x\n",retval);
      return 0;
    }

    //Analog Gain set
     __CCIRegWrite(handle, isregp->Adr_GainL, RegVal_Gain & 0x00FF);
    if(isregp->Adr_GainH != -1){
      __CCIRegWrite(handle, isregp->Adr_GainH, (RegVal_Gain >> 8) & 0x1 );
    }
    __CCIRegWrite(handle, isregp->Adr_ExpH, RegVal_ExpH);
    __CCIRegWrite(handle, isregp->Adr_ExpL, RegVal_ExpL);
    printf("Gain value = %d\n",RegVal_Gain);
    printf("coarse_integration_time = %d\n",coarse_integration_time);
    printf("Initialize done.\n");
    printf("start streaming\n");
    if((retval = ssp_start_streaming(handle)) != SSP_SUCCESS){
        
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }

    struct timeval tv_start, tv_end;
    struct timezone tz_local;
    gettimeofday(&tv_start, &tz_local);
    
    cvWaitKey(2);

    while(1){ if(stop_process==1) break;}
      
    gettimeofday(&tv_end, &tz_local);

    double start_time = tv_start.tv_sec+(double)tv_start.tv_usec/1000000;
    double end_time = tv_end.tv_sec+(double)tv_end.tv_usec/1000000;
    double fps;
    fps = (double)frame_count/(end_time-start_time);
    printf("FrameRate = %f\n", fps);
    
    printf("FrameCount = %d\n", frame_count);
    printf("stop streaming\n");
    if((retval = ssp_stop_streaming(handle)) != SSP_SUCCESS){
        
      printf("Can not start streaming. %x\n",retval);
      return 0;
    }
    
    printf("Flush Wait.\n");
    ssp_flush_event(handle);
    printf("Flush Done.\n");

    if(ssp_finalize(handle) != SSP_SUCCESS){
      printf("Error during finalization.\n");
      return 0;
    }
    
    cvReleaseImage(&RGBImg);
    cvReleaseImage(&BayerImg);
    cvDestroyWindow("SSP Viewer");

    if(ZoomMode == ZOOM_ON){
      cvReleaseImage(&ZoomImg);
      cvDestroyWindow("Zoom");
    }
    
    printf("Finalize done.\n");fflush(stdout);
    printf("Gain:%d\n", Gain);
    printf("coarse integration time:%d\n", coarse_integration_time);
  }
  return 1;
}
